import os
import pickle as Pickle
from datetime import datetime

from ..util import legistar_id_guid

from .people import PeopleCollector
from .departments import DeptCollector
from .meeting import MeetingCollector
from .item import ItemCollector
from .action import ActionCollector


class PickleCollector(object):
    def __init__(self):
        self.people = PeopleCollector()
        self.depts = DeptCollector()
        self.meeting = MeetingCollector()
        self.item = ItemCollector()
        self.action = ActionCollector()
        self.dir = 'data'

        self._collectors = dict(people=self.people,
                                depts=self.depts,
                                meeting=self.meeting,
                                item=self.item,
                                action=self.action)

    def _collector(self, type):
        return self._collectors[type]

    def _filename(self, type, id=None):
        if type == 'people':
            filename = 'people.pickle'
        elif type == 'depts':
            filename = 'departments.pickle'
        elif type == 'meeting':
            filename = 'meeting-%d.pickle' % id
        elif type == 'item':
            filename = 'item-%d.pickle' % id
        elif type == 'action':
            filename = 'action-%d.pickle' % id
        else:
            raise RuntimeError('unknown type')
        return os.path.join(self.dir, filename)

    def get_filename(self, type, id):
        return self._filename(type, id)

    def _dbname(self, type, id=None):
        if type == 'people':
            dbname = 'people'
        elif type == 'depts':
            dbname = 'departments'
        elif type == 'meeting':
            dbname = 'meeting-%d' % id
        elif type == 'item':
            dbname = 'item-%d' % id
        elif type == 'action':
            dbname = 'action-%d' % id
        else:
            raise RuntimeError('unknown type')
        return dbname

    def make_cache_object(self, type, link=None):
        from ..database import MainCache
        id = None
        if type in ['meeting', 'item', 'action']:
            id, guid = legistar_id_guid(link)
        filename = self._filename(type, id)
        dbname = self._dbname(type, id)
        if os.path.isfile(filename):
            content = Pickle.load(open(filename, 'rb'), encoding='utf-8')
            now = datetime.now()
            mc = MainCache()
            mc.name = dbname
            mc.retrieved = now
            mc.updated = now
            mc.content = content
        else:
            raise RuntimeError("No file present %s" % filename)
        return mc

    def collect(self, type, link=None):
        id = None
        if type in ['meeting', 'item', 'action']:
            id, guid = legistar_id_guid(link)
        filename = self._filename(type, id)
        if not os.path.isfile(filename):
            print("Retrieving %s from legistar..." % filename)
            collector = self._collector(type)
            if link is not None:
                print('link is', link, type)
                if not link.startswith('http'):
                    link = collector.url_prefix + link
                print("Retrieving", link)
                collector.set_url(link)
            collector.collect()
            data = dict(result=collector.result, content=collector.content)
            Pickle.dump(data, open(filename, 'wb'))
        try:
            data = Pickle.load(open(filename, 'rb'))
        except UnicodeDecodeError:
            data = Pickle.load(open(filename, 'rb'), encoding='bytes')
        return data['result']


class _MainCollector(PeopleCollector, DeptCollector,
                     MeetingCollector, ItemCollector,
                     ActionCollector):
    pass


class MainCollector(_MainCollector):
    def __init__(self):
        _MainCollector.__init__(self)
        self.dept_url = 'https://hattiesburg.legistar.com/Departments.aspx'
        self.people_url = 'https://hattiesburg.legistar.com/People.aspx'
        self.url_prefix = 'https://hattiesburg.legistar.com/'
        self._map = dict(people=PeopleCollector,
                         dept=DeptCollector,
                         meeting=MeetingCollector,
                         item=ItemCollector,
                         action=ActionCollector)

    def collect(self, ctype):
        self._map[ctype].collect(self)



import feedparser
import requests


class RssCollector(object):
    def __init__(self):
        self.rss = None
        self.info = None
        self.entries = []

    def get_rss(self, url, content=None):
        if content is None:
            r = requests.get(url)
            self.content = r.content
            self.info = r.header
        else:
            self.content = content
        self.rss = feedparser.parse(self.content)
        self.entries = self.rss.entries



import re

# from ..util import onclick_link
from ..util import legistar_id_guid

from .base import BaseCollector

SPANIDS = dict(agenda_status='_lblAgendaStatus',
               minutes_status='_lblMinutesStatus',
               date='_lblDate',
               time='_lblTime')

ANCHORIDS = dict(agenda='_hypAgenda',
                 minutes='_hypMinutes',
                 dept='_hypName')


MEETING_ITEM_COLUMNS = ['file_id',
                        'version',
                        'agenda_num',
                        'name',
                        'type',
                        'title',
                        'action',
                        'result',
                        'action_details',
                        'video'
                        ]


class MeetingCollector(BaseCollector):
    def _get_meeting_info(self, page):
        info = {'anchors': {},
                'spans': {}
                }
        for key in SPANIDS:
            exp = re.compile('.+%s$' % SPANIDS[key])
            spans = page.find_all('span', id=exp)
            if len(spans) != 1:
                print("No span for %s" % key)
            info['spans'][key] = spans[0]
        for key in ANCHORIDS:
            exp = re.compile('.+%s$' % ANCHORIDS[key])
            anchors = page.find_all('a', id=exp)
            if len(anchors) != 1:
                raise RuntimeError("No anchor for %s" % key)
            info['anchors'][key] = anchors[0]
        return info

    def _prepare_meeting_info(self, info):
        newinfo = {}
        for key in list(info['anchors'].keys()):
            if 'href' in info['anchors'][key].attrs:
                newinfo[key] = info['anchors'][key]['href']
            else:
                newinfo[key] = None
        for key in list(info['spans'].keys()):
            newinfo[key] = info['spans'][key].text
        return newinfo

    def get_meeting_info(self, page):
        info = self._get_meeting_info(page)
        return self._prepare_meeting_info(info)

    def _make_meeting_item(self, row):
        colnames = MEETING_ITEM_COLUMNS
        cols = row.find_all('td')
        item = {}.fromkeys(colnames)
        item['item_page'] = cols[0].a['href']
        # generic parsing from file_id to title
        for index in range(len(colnames)):
            text = cols[index].text.strip()
            if text:
                item[colnames[index]] = text
        return item

    def get_meeting_items(self, page):
        tables = page.find_all('table', class_='rgMasterTable')
        if len(tables) > 1:
            msg = "Problem with determining master table len(tables) = %d"
            raise RuntimeError(msg % len(tables))
        elif not tables:
            return []
        table = tables.pop()
        items = []
        # skip the head row
        for row in table.find_all('tr')[1:]:
            # skip a table with no records
            if 'class' in row.attrs and row.attrs['class'] == ['rgNoRecords']:
                continue
            item = self._make_meeting_item(row)
            items.append(item)
        return items

    def get_meeting(self, page):
        items = self.get_meeting_items(page)
        info = self.get_meeting_info(page)
        # FIXME: THIS DEPENDS ON USING collect()
        info['id'], info['guid'] = legistar_id_guid(self.url)
        info['link'] = self.url
        id, guid = legistar_id_guid(info['dept'])
        info['dept_id'] = id
        return dict(items=items, info=info)

    def collect(self):
        self.retrieve_page(url=self.url)
        self.meeting = self.get_meeting(self.soup)
        self.result = self.meeting



from datetime import datetime

from ..database import AgendaItemTypeMap


def convert_range_to_datetime(start, end):
    "start and end are timestamps"
    start = datetime.fromtimestamp(float(start))
    end = datetime.fromtimestamp(float(end))
    return start, end


def convert_agenda_number(agenda_number):
    delimiter = '.-'
    for delimiter in ['.-', '. - ', ' - ', '-']:
        if delimiter in agenda_number:
            break
    if delimiter in agenda_number:
        itemtype, order = agenda_number.split(delimiter)
        itemtype = AgendaItemTypeMap[itemtype]
        order = int(order)
    else:
        itemtype = 'unknown'
        if agenda_number:
            order = int(agenda_number)
        else:
            order = None
    return itemtype, order



import robobrowser


class BaseCollector(object):
    def __init__(self):
        self.browser = robobrowser.RoboBrowser(parser='lxml')
        self.url = None
        self.response = None
        self.pageinfo = None
        self.content = ''
        self.soup = None
        self.url_prefix = 'https://hattiesburg.legistar.com/'

    def retrieve_page(self, url=None):
        if url is None:
            url = self.url
        else:
            self.url = url
        if url is None:
            raise RuntimeError("No url set.")
        self.response = self.browser.open(url)
        self.info = self.browser.response.headers
        self.content = self.browser.response.content
        self.soup = self.browser.parsed

    def set_url(self, url):
        self.url = url
        self.response = None
        self.pageinfo = None
        self.content = ''
        self.soup = None

    def collect(self):
        pass



import re

from ..util import legistar_id_guid

from .base import BaseCollector

ACTION_DATA_IDENTIFIERS = dict(file_id='_hypFile',
                               ftype='_lblType',
                               mover='_hypMover',
                               seconder='_hypSeconder',
                               result='_lblResult',
                               agenda_note='_lblAgendaNote',
                               minutes_note='_lblMinutesNote',
                               action='_lblAction',
                               action_text='_lblActionText')


class ActionCollector(BaseCollector):
    def _get_votes(self, page):
        tables = page.find_all('table', class_='rgMasterTable')
        if len(tables) != 1:
            msg = "Problem with determining master table len(tables) = %d"
            raise RuntimeError(msg % len(tables))
        table = tables.pop()
        items = []
        for row in table.find_all('tr')[1:]:
            if 'rgNoRecords' in row['class']:
                return []
            person, vote = row.find_all('td')
            name = person.text.strip()
            link = person.a['href']
            vote = vote.text.strip()
            items.append((name, link, vote))
        return items

    def _get_action(self, page):
        markers = ACTION_DATA_IDENTIFIERS
        item_keys = list(markers.keys())
        item = {}.fromkeys(item_keys)
        item['roll_call'] = False
        found_tags = False
        for key in item_keys:
            exp = re.compile('.+%s$' % markers[key])
            tags = page.find_all('span', id=exp)
            if not tags:
                tags = page.find_all('a', id=exp)
            if not tags:
                print("NO TAGS FOR", key)
                continue
            found_tags = True
            if len(tags) > 1:
                print("len(%s) == %d" % (key, len(tags)))
            tag = tags[0]
            if markers[key].startswith('_lbl'):
                # just text with lbl
                item[key] = tag.text.strip()
            elif markers[key].startswith('_hyp'):
                name = tag.text.strip()
                link = tag['href']
                item[key] = (name, link)
            else:
                item[key] = tag
        if not found_tags:
            rollcall_id = 'ctl00_ContentPlaceHolder1_pageRollCall'
            if page.find_all('div', id=rollcall_id):
                item['roll_call'] = True
        item['votes'] = self._get_votes(page)
        return item

    def collect(self):
        self.retrieve_page(self.url)
        self.action = self._get_action(self.soup)
        self.action['id'], self.action['guid'] = legistar_id_guid(self.url)
        self.result = self.action
